from jax import core, numpy as jnp
from jax.extend.core import Primitive
from onnx import helper
from typing import TYPE_CHECKING, Union, Sequence, Tuple
from jax2onnx.plugin_system import register_primitive, PrimitivePlugin

if TYPE_CHECKING:
    from jax2onnx.converter.converter import Jaxpr2OnnxConverter

import numpy as np

# Define a new primitive for tile
jnp.tile_p = Primitive("jnp.tile")
jnp.tile_p.multiple_results = False


@register_primitive(
    jaxpr_primitive=jnp.tile_p.name,
    jax_doc="https://jax.readthedocs.io/en/latest/_autosummary/jax.numpy.tile.html",
    onnx=[
        {
            "component": "Tile",
            "doc": "https://onnx.ai/onnx/operators/onnx__Tile.html",
        }
    ],
    since="v0.1.0",
    context="primitives.jnp",
    testcases=[
        {
            "testcase": "tile_a",
            "callable": lambda a: jnp.tile(a, (1, 2)),
            "input_shapes": [(2, 3)],
        },
        {
            "testcase": "tile_b",
            "callable": lambda a: jnp.tile(a, (1, 2, 1)),
            "input_shapes": [(1, 5, 5)],
        },
        {
            "testcase": "tile_c",
            "callable": lambda a: jnp.tile(a, (1, 4)),
            "input_shapes": [(3, 3)],
        },
        {
            "testcase": "tile_d",
            "callable": lambda a: jnp.tile(a, 2),  # Scalar reps
            "input_shapes": [(3, 3)],
        },
        {
            "testcase": "tile_dynamic",
            "callable": lambda a: jnp.tile(a, (2, 1)),  # Repeat batch dimension
            "input_shapes": [("B", 3)],
        },
        {  # Test case to check padding.
            "testcase": "tile_pad",
            "callable": lambda a: jnp.tile(a, (2, 3, 4)),
            "input_shapes": [(4, 5)],
        },
    ],
)
class TilePlugin(PrimitivePlugin):
    """
    Plugin for converting jax.numpy.tile to ONNX.
    """

    @staticmethod
    def _tile_abstract_eval(x, repeats: Sequence[int]):
        """Compute the output shape for tile."""
        x_shape = x.shape
        if len(repeats) != len(x_shape):
            if len(repeats) < len(x_shape):
                repeats = (1,) * (len(x_shape) - len(repeats)) + tuple(repeats)
            else:
                x_shape = (1,) * (len(repeats) - len(x_shape)) + x_shape

        output_shape = tuple(s * r for s, r in zip(x_shape, repeats))
        return core.ShapedArray(output_shape, x.dtype)

    @staticmethod
    def abstract_eval(x, repeats: Sequence[int]):
        return TilePlugin._tile_abstract_eval(x, repeats)

    def to_onnx(self, s: "Jaxpr2OnnxConverter", node_inputs, node_outputs, params):
        """Handles ONNX conversion for jnp.tile."""
        repeats = params["repeats"]
        input_name = s.get_name(node_inputs[0])
        output_name = s.get_name(node_outputs[0])
        input_shape = node_inputs[0].aval.shape

        # ONNX requires repeats as an initializer.
        repeats_name = s.get_unique_name("tile_repeats")

        # If repeats has more dimensions than input, reshape input first
        actual_input_name = input_name
        if len(repeats) > len(input_shape):
            # Add leading dimensions of size 1
            reshaped_input_shape = (1,) * (
                len(repeats) - len(input_shape)
            ) + input_shape

            # Create shape tensor for reshaping
            shape_name = s.get_unique_name("reshape_shape")
            s.add_initializer(
                name=shape_name, vals=np.array(reshaped_input_shape, dtype=np.int64)
            )

            # Create reshape node
            reshaped_name = s.get_unique_name("reshaped_input")
            reshape_node = helper.make_node(
                "Reshape",
                inputs=[input_name, shape_name],
                outputs=[reshaped_name],
                name=s.get_unique_name("reshape"),
                allowzero=0,  # Explicit allowzero
            )
            s.add_node(reshape_node)
            s.add_shape_info(reshaped_name, reshaped_input_shape)  # Add shape info

            # Update input name and shape for tile operation
            actual_input_name = reshaped_name
            input_shape = reshaped_input_shape
        elif len(repeats) < len(input_shape):
            # Pad repeats to match input rank if needed, prepending 1s
            repeats = (1,) * (len(input_shape) - len(repeats)) + tuple(repeats)

        # Add repeats as initializer
        s.add_initializer(name=repeats_name, vals=np.array(repeats, dtype=np.int64))

        # Create tile node
        tile_node = helper.make_node(
            "Tile",
            inputs=[actual_input_name, repeats_name],
            outputs=[output_name],
            name=s.get_unique_name("tile"),
        )
        s.add_node(tile_node)

        # Calculate output shape
        output_shape = tuple(s * r for s, r in zip(input_shape, repeats))
        s.add_shape_info(output_name, output_shape)

    @staticmethod
    def _tile(a, reps: Union[int, Sequence[int]]):
        """Defines the primitive binding for Tile."""
        reps_tuple = TilePlugin._determine_dimensions(reps, len(a.shape))
        return jnp.tile_p.bind(a, repeats=reps_tuple)

    @staticmethod
    def _determine_dimensions(
        reps: Union[int, Sequence[int]], operand_ndim: int
    ) -> Tuple[int, ...]:
        """Handle different representations of repetition dimensions."""
        if isinstance(reps, int):
            reps_tuple: Tuple[int, ...] = (reps,)
        else:
            reps_tuple = tuple(reps)
        if len(reps_tuple) < operand_ndim:
            reps_tuple = (1,) * (operand_ndim - len(reps_tuple)) + reps_tuple
        return reps_tuple

    @staticmethod
    def get_monkey_patch():
        """Provides patching information for Tile."""

        def patched_tile(a, reps: Union[int, Sequence[int]]):
            return TilePlugin._tile(a, reps)

        return patched_tile

    @staticmethod
    def patch_info():
        """Provides patching information for Tile."""
        return {
            "patch_targets": [jnp],
            "patch_function": lambda _: TilePlugin._tile,
            "target_attribute": "tile",
        }


# Register abstract evaluation function
jnp.tile_p.def_abstract_eval(TilePlugin.abstract_eval)



"""
Conv Transpose Plugin for JAX to ONNX conversion.

This plugin enables conversion of flax.nnx.ConvTranspose layers to ONNX format.
It transforms JAX’s conv_transpose operations into an ONNX ConvTranspose operator
with necessary Transpose operations for NHWC to NCHW conversion.
"""

from typing import TYPE_CHECKING
from flax import nnx
from jax import core
from onnx import helper
from jax.extend.core import Primitive
from jax2onnx.plugin_system import register_primitive, PrimitivePlugin

if TYPE_CHECKING:
    from jax2onnx.converter.converter import Jaxpr2OnnxConverter

# Define a new primitive for conv transpose.
nnx.conv_transpose_p = Primitive("nnx.conv_transpose")
nnx.conv_transpose_p.multiple_results = False  # Set once at initialization


def _convert_padding(padding):
    if isinstance(padding, str):
        p = padding.upper()
        if p == "VALID":
            return [0, 0, 0, 0]
        elif p == "SAME":
            return [1, 1, 1, 1]
        elif p == "CIRCULAR":
            return [2, 2, 2, 2]
    return padding


@register_primitive(
    jaxpr_primitive=nnx.conv_transpose_p.name,
    jax_doc="https://flax.readthedocs.io/en/latest/api_reference/flax.nnx/nn/conv_transpose.html",
    onnx=[
        {
            "component": "ConvTranspose",
            "doc": "https://onnx.ai/onnx/operators/onnx__ConvTranspose.html",
        },
    ],
    since="v0.3.0",
    context="primitives.nnx",
    testcases=[
        {
            "testcase": "conv_transpose_valid_padding",
            "callable": nnx.ConvTranspose(
                in_features=3,
                out_features=4,
                kernel_size=(3,),
                padding="VALID",
                rngs=nnx.Rngs(0),
            ),
            "input_shapes": [(1, 8, 3)],
        },
        {
            "testcase": "conv_transpose_circular_padding",
            "callable": nnx.ConvTranspose(
                in_features=3,
                out_features=4,
                kernel_size=(6, 6),
                strides=(2, 2),
                padding="CIRCULAR",
                transpose_kernel=True,
                rngs=nnx.Rngs(0),
            ),
            "input_shapes": [(1, 15, 15, 3)],
        },
    ],
)
class ConvTransposePlugin(PrimitivePlugin):

    @staticmethod
    def abstract_eval(x, weight, *args, **kwargs):
        pads_raw = kwargs.get("pads", [0, 0, 0, 0])
        is_circular = isinstance(pads_raw, str) and pads_raw.upper() == "CIRCULAR"

        strides = kwargs.get("strides", (1, 1))
        pads = kwargs.get("pads", [0, 0, 0, 0])
        if isinstance(pads, str):
            pads = _convert_padding(pads)
        dilations = kwargs.get("dilations", (1, 1))
        output_padding = kwargs.get("output_padding", (0, 0))

        if len(x.shape) == 3:
            output_shape = ConvTransposePlugin.calculate_output_shape_1d(
                x.shape, weight.shape, strides, pads, dilations, output_padding
            )
        elif len(x.shape) == 4:
            output_shape = ConvTransposePlugin.calculate_output_shape_2d(
                x.shape, weight.shape, strides, pads, dilations, output_padding
            )
            if is_circular:
                # For circular mode, JAX produces an extra singleton dimension after batch.
                output_shape = (output_shape[0], 1) + output_shape[1:]
        else:
            raise ValueError(f"Unexpected input shape: {x.shape}")

        return core.ShapedArray(output_shape, x.dtype)

    @staticmethod
    def calculate_output_shape_2d(
        input_shape, weight_shape, strides, pads, dilations, output_padding
    ):
        batch_size, input_height, input_width, in_channels = input_shape
        kernel_height, kernel_width, _, out_channels = weight_shape

        if isinstance(pads, str):
            p = pads.upper()
            if p == "VALID":
                pads = [0, 0, 0, 0]
            elif p == "SAME":
                pad_along_height = max(
                    (input_height - 1) * strides[0]
                    + dilations[0] * (kernel_height - 1)
                    + 1
                    - input_height,
                    0,
                )
                pad_along_width = max(
                    (input_width - 1) * strides[1]
                    + dilations[1] * (kernel_width - 1)
                    + 1
                    - input_width,
                    0,
                )
                pad_top = pad_along_height // 2
                pad_bottom = pad_along_height - pad_top
                pad_left = pad_along_width // 2
                pad_right = pad_along_width - pad_left
                pads = [pad_top, pad_left, pad_bottom, pad_right]
            elif p == "CIRCULAR":
                pads = [2, 2, 2, 2]
        elif isinstance(pads, int):
            pads = [pads, pads, pads, pads]
        elif isinstance(pads, tuple):
            if len(pads) == 2:
                pads = [pads[0], pads[1], pads[0], pads[1]]
            elif len(pads) == 4:
                pads = list(pads)
            else:
                raise ValueError(
                    "Unsupported pads tuple length for 2D convolution transpose."
                )

        if isinstance(dilations, int):
            dilations = [dilations, dilations]
        if isinstance(output_padding, int):
            output_padding = [output_padding, output_padding]
        if isinstance(strides, int):
            strides = [strides, strides]
        elif strides is None:
            strides = [1, 1]

        output_height = (
            (input_height - 1) * strides[0]
            - pads[0]
            - pads[2]
            + dilations[0] * (kernel_height - 1)
            + output_padding[0]
            + 1
        )
        output_width = (
            (input_width - 1) * strides[1]
            - pads[1]
            - pads[3]
            + dilations[1] * (kernel_width - 1)
            + output_padding[1]
            + 1
        )

        return (batch_size, output_height, output_width, out_channels)

    @staticmethod
    def calculate_output_shape_1d(
        input_shape, weight_shape, strides, pads, dilations, output_padding
    ):
        batch_size, in_channels, seq_len = input_shape
        kernel_size, _, out_channels = weight_shape

        if isinstance(pads, str):
            if pads.upper() == "VALID":
                pads = [0, 0]
            elif pads.upper() == "SAME":
                pad_along_length = max(
                    (seq_len - 1) * strides[0]
                    + dilations[0] * (kernel_size - 1)
                    + 1
                    - seq_len,
                    0,
                )
                pad_l = pad_along_length // 2
                pad_r = pad_along_length - pad_l
                pads = [pad_l, pad_r]
            elif pads.upper() == "CIRCULAR":
                pads = [1, 1]
        elif isinstance(pads, int):
            pads = [pads, pads]
        elif isinstance(pads, tuple):
            if len(pads) == 1:
                pads = [pads[0], pads[0]]
            else:
                pads = [pads[0], pads[1]]

        if isinstance(dilations, int):
            dilations = [dilations]
        if isinstance(output_padding, int):
            output_padding = [output_padding]
        if isinstance(strides, int):
            strides = [strides]
        elif strides is None:
            strides = [1]

        output_length = (
            (seq_len - 1) * strides[0]
            - 2 * pads[0]
            + dilations[0] * (kernel_size - 1)
            + output_padding[0]
            + 1
        )
        return (batch_size, out_channels, output_length)

    def to_onnx(self, s: "Jaxpr2OnnxConverter", node_inputs, node_outputs, params):
        input_name = s.get_name(node_inputs[0])
        weight_name = s.get_name(node_inputs[1])
        has_bias = len(node_inputs) > 2
        if has_bias:
            bias_name = s.get_name(node_inputs[2])
        final_output_name = s.get_name(node_outputs[0])

        strides = params.get("strides", (1, 1))
        pads = params.get("pads", [0, 0, 0, 0])
        if isinstance(pads, str):
            pads = _convert_padding(pads)
        dilations = params.get("dilations", (1, 1))
        group = params.get("group", 1)
        output_padding = params.get("output_padding", (0, 0))
        transpose_kernel = params.get("transpose_kernel", False)
        is_circular = (
            isinstance(params.get("pads"), str)
            and params.get("pads").upper() == "CIRCULAR"
        )

        jax_shape = node_inputs[0].aval.shape
        node_inputs[1].aval.shape

        if len(jax_shape) == 4:
            if isinstance(strides, int):
                strides = (strides, strides)

            # Pre-transpose input: NHWC -> NCHW.
            pre_transpose_name = s.get_unique_name("convtrans_pre_transpose")
            pre_transpose_node = helper.make_node(
                "Transpose",
                inputs=[input_name],
                outputs=[pre_transpose_name],
                name=s.get_unique_name("convtrans_pre_transpose_node"),
                perm=[0, 3, 1, 2],
            )
            s.add_node(pre_transpose_node)
            pre_transposed_shape = (
                jax_shape[0],
                jax_shape[3],
                jax_shape[1],
                jax_shape[2],
            )
            s.add_shape_info(pre_transpose_name, pre_transposed_shape)
            input_to_conv = pre_transpose_name

            # Transpose weight for 2D convolution.
            weight_transpose_name = s.get_unique_name("convtrans_weight_transpose")
            perm = [2, 3, 0, 1] if not transpose_kernel else [3, 2, 0, 1]
            weight_transpose_node = helper.make_node(
                "Transpose",
                inputs=[weight_name],
                outputs=[weight_transpose_name],
                name=s.get_unique_name("convtrans_weight_transpose_node"),
                perm=perm,
            )
            s.add_node(weight_transpose_node)
            weight_name = weight_transpose_name

            conv_output_name = s.get_unique_name("conv_transpose_output")
            conv_inputs = [input_to_conv, weight_name]
            if has_bias:
                conv_inputs.append(bias_name)

            conv_transpose_node = helper.make_node(
                "ConvTranspose",
                inputs=conv_inputs,
                outputs=[conv_output_name],
                name=s.get_unique_name("conv_transpose"),
                strides=strides,
                pads=pads,
                dilations=dilations,
                group=group,
                output_padding=output_padding,
            )
            s.add_node(conv_transpose_node)

            # Post-transpose output: NCHW -> NHWC.
            s.get_unique_name("convtrans_post_transpose")
            post_transpose_node = helper.make_node(
                "Transpose",
                inputs=[conv_output_name],
                outputs=[final_output_name],
                name=s.get_unique_name("convtrans_post_transpose"),
                perm=[0, 2, 3, 1],
            )
            s.add_node(post_transpose_node)

            if is_circular:
                # Insert an Unsqueeze to add the extra singleton dimension.
                s.get_unique_name("convtrans_unsqueeze")
                unsqueeze_node = helper.make_node(
                    "Unsqueeze",
                    inputs=[final_output_name],
                    outputs=[final_output_name],
                    axes=[1],
                )
                s.add_node(unsqueeze_node)

        elif len(jax_shape) == 3:
            # 1D ConvTranspose branch.
            # Transpose input: (batch, length, channels) -> (batch, channels, length)
            transpose_name = s.get_unique_name("convtrans_input_transpose")
            transpose_node = helper.make_node(
                "Transpose",
                inputs=[input_name],
                outputs=[transpose_name],
                name=s.get_unique_name("convtrans_input_transpose_node"),
                perm=[0, 2, 1],
            )
            s.add_node(transpose_node)
            transposed_shape = (jax_shape[0], jax_shape[2], jax_shape[1])
            s.add_shape_info(transpose_name, transposed_shape)
            input_to_conv = transpose_name

            # Transpose weight for 1D convolution.
            weight_transpose_name = s.get_unique_name("convtrans_weight_transpose")
            perm = [1, 2, 0] if not transpose_kernel else [2, 1, 0]
            weight_transpose_node = helper.make_node(
                "Transpose",
                inputs=[weight_name],
                outputs=[weight_transpose_name],
                name=s.get_unique_name("convtrans_weight_transpose_node"),
                perm=perm,
            )
            s.add_node(weight_transpose_node)
            weight_name = weight_transpose_name

            if strides is None:
                strides = (1,)
            elif isinstance(strides, int):
                strides = (strides,)
            if len(strides) == 1:
                strides = (strides[0],)  # 1D strides remain as a 1-tuple.

            conv_output_name = s.get_unique_name("conv_transpose_output")
            conv_inputs = [input_to_conv, weight_name]
            if has_bias:
                conv_inputs.append(bias_name)

            conv_transpose_node = helper.make_node(
                "ConvTranspose",
                inputs=conv_inputs,
                outputs=[conv_output_name],
                name=s.get_unique_name("conv_transpose"),
                strides=strides,
                pads=pads,
                dilations=dilations,
                group=group,
                output_padding=output_padding,
            )
            s.add_node(conv_transpose_node)

            # Post-transpose output: (batch, channels, length) -> (batch, length, channels)
            s.get_unique_name("convtrans_output_transpose")
            post_transpose_node = helper.make_node(
                "Transpose",
                inputs=[conv_output_name],
                outputs=[final_output_name],
                name=s.get_unique_name("convtrans_output_transpose"),
                perm=[0, 2, 1],
            )
            s.add_node(post_transpose_node)

        else:
            raise ValueError(f"Unsupported input shape for ConvTranspose: {jax_shape}")

    @staticmethod
    def _conv_transpose(
        x,
        weight,
        bias=None,
        strides=(1, 1),
        pads=(0, 0, 0, 0),
        dilations=(1, 1),
        group=1,
        output_padding=(0, 0),
        transpose_kernel=False,
    ):
        nnx.conv_transpose_p.multiple_results = False
        if bias is not None:
            return nnx.conv_transpose_p.bind(
                x,
                weight,
                bias,
                strides=strides,
                pads=pads,
                dilations=dilations,
                group=group,
                output_padding=output_padding,
                transpose_kernel=transpose_kernel,
            )
        else:
            return nnx.conv_transpose_p.bind(
                x,
                weight,
                strides=strides,
                pads=pads,
                dilations=dilations,
                group=group,
                output_padding=output_padding,
                transpose_kernel=transpose_kernel,
            )

    @staticmethod
    def conv_transpose(
        x,
        weight,
        bias,
        strides,
        pads,
        dilations,
        group,
        output_padding,
        transpose_kernel,
    ):
        return ConvTransposePlugin._conv_transpose(
            x,
            weight,
            bias,
            strides,
            pads,
            dilations,
            group,
            output_padding,
            transpose_kernel,
        )

    @staticmethod
    def get_monkey_patch():
        def patched_conv_transpose_call(self, x):
            return ConvTransposePlugin._conv_transpose(
                x,
                self.kernel.value,
                (
                    self.bias.value
                    if hasattr(self, "bias") and self.bias is not None
                    else None
                ),
                strides=self.strides,
                pads=_convert_padding(self.padding),
                dilations=getattr(self, "dilations", (1, 1)),
                group=getattr(self, "group", 1),
                output_padding=getattr(self, "output_padding", (0, 0)),
                transpose_kernel=getattr(self, "transpose_kernel", False),
            )

        return patched_conv_transpose_call

    @staticmethod
    def patch_info():
        return {
            "patch_targets": [nnx.ConvTranspose],
            "patch_function": lambda _: ConvTransposePlugin.get_monkey_patch(),
            "target_attribute": "__call__",
        }


# Register abstract evaluation function.
nnx.conv_transpose_p.def_abstract_eval(ConvTransposePlugin.abstract_eval)



import jax
from typing import TYPE_CHECKING
from onnx import helper
from jax2onnx.plugin_system import register_primitive, PrimitivePlugin

if TYPE_CHECKING:
    from jax2onnx.converter.converter import Jaxpr2OnnxConverter


@register_primitive(
    jaxpr_primitive=jax.lax.concatenate_p.name,
    jax_doc="https://docs.jax.dev/en/latest/_autosummary/jax.lax.concatenate.html",
    onnx=[
        {
            "component": "Concat",
            "doc": "https://onnx.ai/onnx/operators/onnx__Concat.html",
        }
    ],
    since="v0.2.0",
    context="primitives.lax",
    testcases=[
        {
            "testcase": "concatenate",
            "callable": lambda a, b: jax.lax.concatenate(
                (a, b), dimension=0
            ),  # Corrected callable
            "input_shapes": [(3,), (3,)],
        },
        {
            "testcase": "concatenate_axis1",
            "callable": lambda a, b: jax.lax.concatenate(
                (a, b), dimension=1
            ),  # Corrected callable
            "input_shapes": [(2, 3), (2, 4)],
        },
        {
            "testcase": "concatenate_dynamic",
            "callable": lambda a, b: jax.lax.concatenate(
                (a, b), dimension=0
            ),  # Corrected callable
            "input_shapes": [("B", 3), ("B", 3)],
        },
        {  # 3D inputs
            "testcase": "concatenate_3d",
            "callable": lambda a, b: jax.lax.concatenate(
                (a, b), dimension=1
            ),  # Corrected callable
            "input_shapes": [(2, 3, 4), (2, 5, 4)],
        },
    ],
)
class ConcatenatePlugin(PrimitivePlugin):
    """
    Plugin for converting jax.lax.concatenate to ONNX.
    """

    def to_onnx(self, s: "Jaxpr2OnnxConverter", node_inputs, node_outputs, params):
        """Handle JAX concatenate primitive."""
        input_names = [s.get_name(inp) for inp in node_inputs]
        output_name = s.get_name(node_outputs[0])
        dimension = params["dimension"]

        # Calculate and propagate shape information
        input_shapes = [inp.aval.shape for inp in node_inputs]
        output_shape = list(input_shapes[0])  # Start with the shape of the first input
        # Normalize the axis
        rank = len(output_shape)
        dimension = dimension % rank if dimension < 0 else dimension

        for shape in input_shapes[1:]:
            if len(shape) != rank:
                raise ValueError("Inputs must have the same rank.")
            for i in range(rank):
                if i != dimension:
                    if output_shape[i] != shape[i] and not (
                        isinstance(output_shape[i], str) or isinstance(shape[i], str)
                    ):
                        raise ValueError(
                            f"Shapes are incompatible along dimension {i} for Concatenate: {output_shape} vs {shape}"
                        )
                    else:
                        if isinstance(shape[i], str):
                            output_shape[i] = shape[i]

        # Accumulate on the concat axis
        for shape in input_shapes[1:]:
            output_shape[dimension] = (
                output_shape[dimension] + shape[dimension]
                if isinstance(output_shape[dimension], int)
                and isinstance(shape[dimension], int)
                else f"{output_shape[dimension]}+{shape[dimension]}"  # Keep strings
            )

        node = helper.make_node(
            "Concat",
            inputs=input_names,
            outputs=[output_name],
            name=s.get_unique_name("concat"),
            axis=dimension,
        )
        s.add_node(node)
        s.add_shape_info(output_name, tuple(output_shape))



import jax
from typing import TYPE_CHECKING
from onnx import helper
from jax2onnx.plugin_system import register_primitive, PrimitivePlugin

if TYPE_CHECKING:
    from jax2onnx.converter.converter import Jaxpr2OnnxConverter


@register_primitive(
    jaxpr_primitive=jax.lax.log_p.name,
    jax_doc="https://docs.jax.dev/en/latest/_autosummary/jax.lax.log.html",
    onnx=[
        {
            "component": "Log",
            "doc": "https://onnx.ai/onnx/operators/onnx__Log.html",
        }
    ],
    since="v0.2.0",
    context="primitives.lax",
    testcases=[
        {
            "testcase": "log",
            "callable": lambda x: jax.lax.log(x),
            "input_shapes": [(3,)],
        }
    ],
)
class LogPlugin(PrimitivePlugin):
    """Plugin for converting jax.lax.log to ONNX Log."""

    def to_onnx(self, s: "Jaxpr2OnnxConverter", node_inputs, node_outputs, params):
        """Handle JAX log primitive."""
        input_name = s.get_name(node_inputs[0])
        output_name = s.get_var_name(node_outputs[0])
        node = helper.make_node(
            "Log",
            inputs=[input_name],
            outputs=[output_name],
            name=s.get_unique_name("log"),
        )
        s.add_node(node)



"""
Batch Norm Plugin for JAX to ONNX conversion.

This plugin enables conversion of flax.nnx.BatchNorm layers to ONNX format.
It transforms JAX’s batch_norm operations into an ONNX BatchNormalization operator
with necessary Transpose operations for NHWC to NCHW conversion.

The conversion process involves:
  1. Handling input shape and transpositions.
  2. Providing an abstract evaluation for JAX's tracing system.
  3. Converting the operation to ONNX using BatchNormalization and Transpose nodes.
  4. Monkey-patching BatchNorm.__call__ to redirect calls to our primitive.
"""

from typing import TYPE_CHECKING
from flax import nnx
from jax import core
from onnx import helper
from jax.extend.core import Primitive
from jax2onnx.plugin_system import register_primitive, PrimitivePlugin
import jax.numpy as jnp

if TYPE_CHECKING:
    from jax2onnx.converter.converter import Jaxpr2OnnxConverter

# Define a new primitive for batch norm.
nnx.batch_norm_p = Primitive("nnx.batch_norm")
nnx.batch_norm_p.multiple_results = False  # Set once at initialization


@register_primitive(
    jaxpr_primitive=nnx.batch_norm_p.name,
    jax_doc="https://flax.readthedocs.io/en/latest/api_reference/flax.nnx/nn/normalization.html#flax.nnx.BatchNorm",
    onnx=[
        {
            "component": "BatchNormalization",
            "doc": "https://onnx.ai/onnx/operators/onnx__BatchNormalization.html",
        },
    ],
    since="v0.1.0",
    context="primitives.nnx",
    testcases=[
        {
            "testcase": "batch_norm",
            "callable": nnx.BatchNorm(
                num_features=64, epsilon=1e-5, momentum=0.9, rngs=nnx.Rngs(0)
            ),
            "input_shapes": [(11, 2, 2, 64)],
        },
        {
            "testcase": "batch_norm_2",
            "callable": nnx.BatchNorm(num_features=20, rngs=nnx.Rngs(0)),
            "input_shapes": [(2, 20)],
        },
        {
            "testcase": "batch_norm_3d",
            "callable": nnx.BatchNorm(num_features=32, rngs=nnx.Rngs(0)),
            "input_shapes": [(8, 32, 32)],
        },
        {
            "testcase": "batch_norm_float64",
            "callable": nnx.BatchNorm(num_features=64, rngs=nnx.Rngs(0)),
            "input_shapes": [(11, 2, 2, 64)],
            "input_dtype": jnp.float64,
        },
        {
            "testcase": "batch_norm_single_batch",
            "callable": nnx.BatchNorm(num_features=64, rngs=nnx.Rngs(0)),
            "input_shapes": [(1, 2, 2, 64)],
        },
    ],
)
class BatchNormPlugin(PrimitivePlugin):
    """
    Plugin for converting flax.nnx.BatchNorm to ONNX.

    Converts a BatchNorm operation into a BatchNormalization operator
    with necessary Transpose operations for NHWC to NCHW conversion.
    """

    @staticmethod
    def abstract_eval(x, scale, bias, mean, var, *args, **kwargs):
        """Abstract evaluation function for batch_norm."""
        return core.ShapedArray(x.shape, x.dtype)

    def to_onnx(self, s: "Jaxpr2OnnxConverter", node_inputs, node_outputs, params):
        """Handles conversion of batch_norm to ONNX format."""
        input_name = s.get_name(node_inputs[0])
        scale_name = s.get_name(node_inputs[1])
        bias_name = s.get_name(node_inputs[2])
        mean_name = s.get_name(node_inputs[3])
        variance_name = s.get_name(node_inputs[4])
        final_output_name = s.get_name(node_outputs[0])
        epsilon = params.get("epsilon")

        jax_shape = node_inputs[0].aval.shape  # e.g. (11, 2, 2, 64) or (2,20)

        if len(jax_shape) == 4:
            pre_transpose_name = s.get_unique_name("bn_pre_transpose")
            pre_transpose_node = helper.make_node(
                "Transpose",
                inputs=[input_name],
                outputs=[pre_transpose_name],
                name=s.get_unique_name("bn_transpose_pre"),
                perm=[0, 3, 1, 2],  # NHWC -> NCHW
            )
            s.add_node(pre_transpose_node)
            pre_transposed_shape = (
                jax_shape[0],
                jax_shape[3],
                jax_shape[1],
                jax_shape[2],
            )
            s.add_shape_info(pre_transpose_name, pre_transposed_shape)

            bn_output_name = s.get_unique_name("bn_output")
            batch_norm_node = helper.make_node(
                "BatchNormalization",
                inputs=[
                    pre_transpose_name,
                    scale_name,
                    bias_name,
                    mean_name,
                    variance_name,
                ],
                outputs=[bn_output_name],
                name=s.get_unique_name("batch_norm"),
                epsilon=epsilon,
            )
            s.add_node(batch_norm_node)
            s.add_shape_info(bn_output_name, pre_transposed_shape)

            post_transpose_node = helper.make_node(
                "Transpose",
                inputs=[bn_output_name],
                outputs=[final_output_name],
                name=s.get_unique_name("bn_transpose_post"),
                perm=[0, 2, 3, 1],  # NCHW -> NHWC
            )
            s.add_node(post_transpose_node)
        else:
            batch_norm_node = helper.make_node(
                "BatchNormalization",
                inputs=[input_name, scale_name, bias_name, mean_name, variance_name],
                outputs=[final_output_name],
                name=s.get_unique_name("batch_norm"),
                epsilon=epsilon,
            )
            s.add_node(batch_norm_node)

    @staticmethod
    def _batch_norm(x, scale, bias, mean, var, epsilon, use_running_average, momentum):
        nnx.batch_norm_p.multiple_results = False
        return nnx.batch_norm_p.bind(
            x,
            scale,
            bias,
            mean,
            var,
            epsilon=epsilon,
            use_running_average=use_running_average,
            momentum=momentum,
        )

    @staticmethod
    def batch_norm(x, scale, bias, mean, var, epsilon, use_running_average, momentum):
        """Binding function for batch_norm."""
        return BatchNormPlugin._batch_norm(
            x, scale, bias, mean, var, epsilon, use_running_average, momentum
        )

    @staticmethod
    def get_monkey_patch():
        """Returns a patched version of BatchNorm.__call__."""

        def patched_batch_norm_call(self, x):
            return BatchNormPlugin._batch_norm(
                x,
                self.scale.value,
                self.bias.value,
                self.mean.value,
                self.var.value,
                epsilon=self.epsilon,
                use_running_average=self.use_running_average,
                momentum=self.momentum,
            )

        return patched_batch_norm_call

    @staticmethod
    def patch_info():
        """Provides patching information."""
        return {
            "patch_targets": [nnx.BatchNorm],
            "patch_function": lambda _: BatchNormPlugin.get_monkey_patch(),
            "target_attribute": "__call__",
        }


# Register abstract evaluation function.
nnx.batch_norm_p.def_abstract_eval(BatchNormPlugin.abstract_eval)



from typing import TYPE_CHECKING

from flax import nnx
from jax.extend.core import Primitive
from onnx import helper
from jax import core
import numpy as np
from jax2onnx.plugin_system import register_primitive, PrimitivePlugin

if TYPE_CHECKING:
    from jax2onnx.converter.converter import Jaxpr2OnnxConverter

# Define the new primitive for dropout.
nnx.dropout_p = Primitive("nnx.dropout")
nnx.dropout_p.multiple_results = False  # Correctly set at initialization


@register_primitive(
    jaxpr_primitive=nnx.dropout_p.name,
    jax_doc="https://flax.readthedocs.io/en/latest/api_reference/flax.nnx/nn/stochastic.html#flax.nnx.Dropout",
    onnx=[
        {
            "component": "Dropout",
            "doc": "https://onnx.ai/onnx/operators/onnx__Dropout.html",
        }
    ],
    since="v0.1.0",
    context="primitives.nnx",
    testcases=[
        {
            "testcase": "dropout_inference",
            "callable": nnx.Dropout(rate=0.5, deterministic=True, rngs=nnx.Rngs(0)),
            "input_shapes": [(5, 10)],
        },
        # comparison with training mode is not possible due to the random nature of dropout
        # {
        #     "testcase": "dropout_training",
        #     "callable": nnx.Dropout(
        #         rate=0.5, deterministic=False, rngs=nnx.Rngs(0)
        #     ),
        #     "input_shapes": [(5, 10)],
        # },
    ],
)
class DropoutPlugin(PrimitivePlugin):
    """
    Plugin for converting flax.nnx.Dropout to ONNX.
    """

    @staticmethod
    def abstract_eval(x, rate, deterministic):
        """Abstract evaluation function for dropout."""
        return core.ShapedArray(x.shape, x.dtype)

    def to_onnx(self, s: "Jaxpr2OnnxConverter", node_inputs, node_outputs, params):
        """Handles conversion of dropout to ONNX format."""
        input_name = s.get_name(node_inputs[0])
        output_name = s.get_name(node_outputs[0])

        # Retrieve dropout parameters
        rate = params.get("rate", 0.0)
        deterministic = params.get("deterministic", True)

        # ONNX Dropout expects:
        # - ratio as a **second input** instead of an attribute in newer versions.
        # - training_mode (bool) in newer versions.  We handle this implicitly.
        dropout_inputs = [input_name]

        if not deterministic:
            ratio_name = s.get_constant_name(np.array(rate, dtype=np.float32))
            dropout_inputs.append(
                ratio_name
            )  # Add ratio as an input instead of an attribute.

        dropout_node = helper.make_node(
            "Dropout",
            inputs=dropout_inputs,
            outputs=[output_name],
            name=s.get_unique_name("dropout"),
        )

        s.add_node(dropout_node)

    @staticmethod
    def _dropout(x, rate, deterministic):
        """Defines the primitive binding for dropout."""
        return nnx.dropout_p.bind(x, rate=rate, deterministic=deterministic)

    @staticmethod
    def get_monkey_patch():
        """Returns a patched version of Dropout's call method."""

        def patched_dropout_call(self, x, deterministic=None):
            det = deterministic if deterministic is not None else self.deterministic
            return DropoutPlugin._dropout(x, self.rate, det)

        return patched_dropout_call

    @staticmethod
    def patch_info():
        """Provides patching information for dropout."""
        return {
            "patch_targets": [nnx.Dropout],
            "patch_function": lambda _: DropoutPlugin.get_monkey_patch(),
            "target_attribute": "__call__",
        }


# Register abstract evaluation function
nnx.dropout_p.def_abstract_eval(DropoutPlugin.abstract_eval)



from jax import core, numpy as jnp
from jax.extend.core import Primitive
from onnx import helper
from typing import TYPE_CHECKING, Tuple, List, Union, Sequence
from jax2onnx.plugin_system import register_primitive, PrimitivePlugin

if TYPE_CHECKING:
    from jax2onnx.converter.converter import Jaxpr2OnnxConverter

import numpy as np

# Define the reshape primitive
jnp.reshape_p = Primitive("jnp.reshape")
jnp.reshape_p.multiple_results = False  # Correct initialization


@register_primitive(
    jaxpr_primitive=jnp.reshape_p.name,
    jax_doc="https://jax.readthedocs.io/en/latest/_autosummary/jax.numpy.reshape.html",
    onnx=[
        {
            "component": "Reshape",
            "doc": "https://onnx.ai/onnx/operators/onnx__Reshape.html",
        }
    ],
    since="v0.1.0",
    context="primitives.jnp",
    testcases=[
        {
            "testcase": "reshape_1",
            "callable": lambda a: jnp.reshape(a, (2, 6)),
            "input_shapes": [(3, 4)],
        },
        {
            "testcase": "reshape_2",
            "callable": lambda a: jnp.reshape(a, (-1, 2)),
            "input_shapes": [(3, 4)],
        },
        {
            "testcase": "reshape_3",
            "callable": lambda a: jnp.reshape(a, (2, -1)),
            "input_shapes": [(3, 4)],
        },
        {
            "testcase": "reshape_4",
            "callable": lambda a: jnp.reshape(a, (-1, 4)),
            "input_shapes": [("B", 3, 4)],
        },
        {
            "testcase": "reshape_to_scalar",
            "callable": lambda a: jnp.reshape(a, ()),
            "input_shapes": [(1,)],
        },
        {
            "testcase": "reshape_from_scalar",
            "callable": lambda a: jnp.reshape(a, (1,)),
            "input_shapes": [()],
        },
    ],
)
class ReshapePlugin(PrimitivePlugin):
    """
    Plugin for converting jax.numpy.reshape to ONNX.
    """

    @staticmethod
    def _process_newshape(newshape: Sequence[Union[int, str]]) -> List[Union[int, str]]:
        """Validates and processes the newshape argument for reshape."""
        if isinstance(newshape, (int, str)):
            newshape = [newshape]
        else:
            newshape = list(newshape)

        neg_one_count = sum(1 for dim in newshape if dim == -1)
        if neg_one_count > 1:
            raise ValueError("Only one dimension can be -1 (inferred).")

        return newshape

    @staticmethod
    def _get_dynamic_output_shape(
        input_shape: Tuple[Union[int, str], ...], newshape: Sequence[Union[int, str]]
    ) -> Tuple[Union[int, str], ...]:
        """Computes the output shape for jnp.reshape while handling dynamic dimensions."""
        newshape = ReshapePlugin._process_newshape(newshape)
        input_shape_list = list(input_shape)

        dummy_input_shape = [1 if isinstance(s, str) else s for s in input_shape_list]
        dummy_newshape = [1 if isinstance(s, str) else s for s in newshape]

        if -1 in dummy_newshape:
            neg_one_index = dummy_newshape.index(-1)
            known_dims_product = np.prod([dim for dim in dummy_newshape if dim != -1])
            # Avoid ZeroDivisionError
            if known_dims_product == 0 and np.prod(dummy_input_shape) != 0:
                raise ValueError(
                    f"Cannot reshape array of shape {input_shape} into shape {newshape}"
                )
            inferred_dim = (
                int(np.prod(dummy_input_shape) / known_dims_product)
                if known_dims_product != 0
                else 0
            )
            dummy_newshape[neg_one_index] = inferred_dim

        if np.prod(dummy_input_shape) != np.prod(dummy_newshape):
            raise ValueError(
                f"Cannot reshape array of shape {input_shape} into shape {newshape}"
            )

        output_shape = [
            orig if isinstance(orig, str) else dummy
            for orig, dummy in zip(newshape, dummy_newshape)
        ]
        return tuple(output_shape)

    @staticmethod
    def abstract_eval(a, newshape):
        """Abstract evaluation function for Reshape."""
        newshape_processed = ReshapePlugin._process_newshape(newshape)
        output_shape = ReshapePlugin._get_dynamic_output_shape(
            a.shape, newshape_processed
        )
        return core.ShapedArray(tuple(output_shape), a.dtype)

    def to_onnx(self, s: "Jaxpr2OnnxConverter", node_inputs, node_outputs, params):
        """Handles conversion of Reshape to ONNX format."""
        newshape = params["newshape"]
        input_name = s.get_name(node_inputs[0])
        output_name = s.get_name(node_outputs[0])

        input_shape = node_inputs[0].aval.shape
        output_shape = ReshapePlugin._get_dynamic_output_shape(input_shape, newshape)
        processed_newshape = ReshapePlugin._process_newshape(newshape)

        # Create a shape tensor for ONNX
        shape_tensor_name = s.get_unique_name("reshape_shape")
        onnx_shape = [dim if isinstance(dim, int) else -1 for dim in processed_newshape]
        # Create a NumPy array with the correct dtype (int64)
        onnx_shape_array = np.array(onnx_shape, dtype=np.int64)
        s.add_initializer(name=shape_tensor_name, vals=onnx_shape_array)

        reshape_node = helper.make_node(
            "Reshape",
            inputs=[input_name, shape_tensor_name],
            outputs=[output_name],
            name=s.get_unique_name("reshape"),
            allowzero=0,  # Explicit allowzero=0
        )
        s.add_node(reshape_node)
        s.add_shape_info(output_name, output_shape)

    @staticmethod
    def _reshape(a, newshape, order="C"):
        """Defines the primitive binding for Reshape."""
        if order != "C":
            raise NotImplementedError("Only C-style reshape is supported.")
        return jnp.reshape_p.bind(a, newshape=newshape)

    @staticmethod
    def get_monkey_patch():
        """Provides patching information for Reshape."""

        def patched_reshape(a, newshape, order="C"):
            return ReshapePlugin._reshape(a, newshape, order)

        return patched_reshape

    @staticmethod
    def patch_info():
        """Provides patching information for Reshape."""
        return {
            "patch_targets": [jnp],
            "patch_function": lambda _: ReshapePlugin.get_monkey_patch(),
            "target_attribute": "reshape",
        }


# Register abstract evaluation function
jnp.reshape_p.def_abstract_eval(ReshapePlugin.abstract_eval)



import jax
import numpy as np
from typing import TYPE_CHECKING
from onnx import helper
from jax2onnx.plugin_system import register_primitive, PrimitivePlugin

if TYPE_CHECKING:
    from jax2onnx.converter.converter import Jaxpr2OnnxConverter


@register_primitive(
    jaxpr_primitive=jax.lax.slice_p.name,
    jax_doc="https://docs.jax.dev/en/latest/_autosummary/jax.lax.slice.html",
    onnx=[
        {
            "component": "Slice",
            "doc": "https://onnx.ai/onnx/operators/onnx__Slice.html",
        }
    ],
    since="v0.1.0",
    context="primitives.lax",
    testcases=[
        {
            "testcase": "slice_test1",
            "callable": lambda x: x[1:3],
            "input_shapes": [(5,)],
        }
    ],
)
class SlicePlugin(PrimitivePlugin):
    """Plugin for converting jax.lax.slice to ONNX Slice."""

    def to_onnx(self, s: "Jaxpr2OnnxConverter", node_inputs, node_outputs, params):
        """Handle JAX slice primitive."""
        input_name = s.get_name(node_inputs[0])
        output_name = s.get_var_name(node_outputs[0])
        start_indices = params["start_indices"]
        limit_indices = params["limit_indices"]
        axes = list(range(len(start_indices)))
        starts_name = s.get_constant_name(np.array(start_indices, dtype=np.int64))
        ends_name = s.get_constant_name(np.array(limit_indices, dtype=np.int64))
        axes_name = s.get_constant_name(np.array(axes, dtype=np.int64))
        inputs_list = [input_name, starts_name, ends_name, axes_name]
        if "strides" in params and params["strides"]:
            strides = params["strides"]
            steps_name = s.get_constant_name(np.array(strides, dtype=np.int64))
            inputs_list.append(steps_name)
        node = helper.make_node(
            "Slice",
            inputs=inputs_list,
            outputs=[output_name],
            name=s.get_unique_name("slice"),
        )
        s.add_node(node)



# file: jax2onnx/examples/autoencoder.py
from flax import nnx
import jax
from jax2onnx.plugin_system import register_example


def Encoder(rngs):
    return nnx.Linear(2, 10, rngs=rngs)


def Decoder(rngs):
    return nnx.Linear(10, 2, rngs=rngs)


class AutoEncoder(nnx.Module):
    def __init__(self, rngs):
        self.encoder = Encoder(rngs)
        self.decoder = Decoder(rngs)

    def __call__(self, x: jax.Array) -> jax.Array:
        return self.decoder(self.encoder(x))

    def encode(self, x: jax.Array) -> jax.Array:
        return self.encoder(x)


register_example(
    component="AutoEncoder",
    description="A simple autoencoder example.",
    source="https://github.com/google/flax/blob/main/README.md",
    since="v0.2.0",
    context="examples.nnx",
    children=["Encoder", "Decoder"],
    testcases=[
        {
            "testcase": "autoencoder",
            "callable": AutoEncoder(rngs=nnx.Rngs(0)),
            "input_shapes": [(1, 2)],
        }
    ],
)



# file: jax2onnx/converter/optimize_transpose.py


import onnx
from onnx import shape_inference
from typing import Dict, List


def remove_redundant_casts(onnx_model: onnx.ModelProto) -> onnx.ModelProto:
    """
    Remove Cast nodes that cast a tensor to its own type.

    This function first runs shape inference to populate type information in the graph,
    then examines each Cast node. If the input tensor's element type is the same as
    the "to" attribute of the Cast node, the node is redundant and is removed,
    with its consumers rewired to the original input.

    Args:
        onnx_model: The input ONNX model.

    Returns:
        The optimized ONNX model with redundant Cast nodes removed.
    """
    # Run shape inference to obtain type information.
    inferred_model = shape_inference.infer_shapes(onnx_model)
    graph = inferred_model.graph

    # Build a mapping from tensor name to its element type.
    type_dict: Dict[str, int] = {}

    def update_type_info(values):
        for value in values:
            # value.type.tensor_type.elem_type is an int corresponding to TensorProto enum.
            type_dict[value.name] = value.type.tensor_type.elem_type

    update_type_info(graph.input)
    update_type_info(graph.value_info)
    update_type_info(graph.output)
    for init in graph.initializer:
        type_dict[init.name] = init.data_type

    nodes_to_remove: List[onnx.NodeProto] = []

    # Iterate over nodes to find redundant Casts.
    for node in graph.node:
        if node.op_type != "Cast":
            continue

        # Get the "to" attribute from the node.
        to_attr = None
        for attr in node.attribute:
            if attr.name == "to":
                to_attr = attr.i
                break
        if to_attr is None:
            continue

        # The Cast node should have one input.
        cast_inp = node.input[0]
        # Check if we know the element type for the input.
        if cast_inp not in type_dict:
            continue

        input_elem_type = type_dict[cast_inp]
        # If the target type equals the input's type, the cast is redundant.
        if input_elem_type != to_attr:
            continue

        # Rewire: for all nodes consuming the output of this Cast, replace with cast_inp.
        cast_out = node.output[0]
        for n in graph.node:
            for idx, inp in enumerate(n.input):
                if inp == cast_out:
                    n.input[idx] = cast_inp

        # Also update graph outputs if needed.
        for out in graph.output:
            if out.name == cast_out:
                out.name = cast_inp

        nodes_to_remove.append(node)

    # Remove the redundant Cast nodes.
    new_nodes = [n for n in graph.node if n not in nodes_to_remove]
    del graph.node[:]
    graph.node.extend(new_nodes)
    return inferred_model


# Define the set of allowed elementwise operations.
ALLOWED_ELEMENTWISE_OPS = {"Elu", "Gelu", "Relu", "Sigmoid", "Tanh"}


def remove_redundant_transpose_pairs(onnx_model: onnx.ModelProto) -> onnx.ModelProto:
    """
    Remove Transpose pairs (possibly separated by elementwise-only nodes)
    whose combined permutation is the identity.

    This function looks for a chain:
      T1 -> [E1 -> E2 -> ... -> E_k] -> T2
    where T1 and T2 are Transpose nodes and E* are elementwise ops (from ALLOWED_ELEMENTWISE_OPS)
    that do not change the ordering of elements.

    When the composed permutation (i.e. T2∘T1) is the identity,
    T1 and T2 are removed and the graph is rewired:
      - For the first elementwise node (if any), its input is replaced with T1's input.
      - Consumers of T2's output are rewired to use the output of the last elementwise node,
        or T1's input if there is no intermediate elementwise op.

    The function modifies the graph in place and returns the modified model.
    """
    graph = onnx_model.graph

    # Build a mapping from tensor name to list of consumer nodes.
    output_to_consumers: Dict[str, List[onnx.NodeProto]] = {}
    for node in graph.node:
        for inp in node.input:
            output_to_consumers.setdefault(inp, []).append(node)

    # Use a list to track nodes to remove.
    nodes_to_remove: List[onnx.NodeProto] = []

    # Iterate over a snapshot of nodes.
    for node in list(graph.node):
        if node in nodes_to_remove:
            continue
        if node.op_type != "Transpose":
            continue

        # Start building the chain with the first Transpose.
        chain = [node]
        current_node = node

        # Walk downstream along the unique-consumer chain.
        while True:
            out_name = current_node.output[0]
            consumers = output_to_consumers.get(out_name, [])
            if len(consumers) != 1:
                break  # Cannot extend chain uniquely.
            next_node = consumers[0]
            # If the next node is one of the allowed elementwise ops, add it to the chain.
            if next_node.op_type in ALLOWED_ELEMENTWISE_OPS:
                chain.append(next_node)
                current_node = next_node
                continue
            # Otherwise, if the next node is a Transpose, add it and stop.
            elif next_node.op_type == "Transpose":
                chain.append(next_node)
            break

        # Only remove if we have at least a pair (chain length >= 2).
        if len(chain) < 2:
            continue

        # At this point, chain = [T1, (E1,...,E_k)*, T2].
        T1 = chain[0]
        T2 = chain[-1]

        # Get permutation attributes from T1 and T2.
        perm_attr1 = [attr for attr in T1.attribute if attr.name == "perm"]
        perm_attr2 = [attr for attr in T2.attribute if attr.name == "perm"]
        if not perm_attr1 or not perm_attr2:
            continue
        perm1 = list(perm_attr1[0].ints)
        perm2 = list(perm_attr2[0].ints)

        # Compose the two permutations: composed[i] = perm1[perm2[i]]
        composed = [perm1[p] for p in perm2]
        # Check if the composed permutation is the identity.
        if composed != list(range(len(composed))):
            continue

        # --- Rewire the graph to bypass T1 and T2 ---
        # For the first node after T1 (if any elementwise op exists), replace its input.
        if len(chain) > 2:
            first_elem_node = chain[1]
            # Replace any occurrence of T1's output with T1's input.
            new_input = T1.input[0]
            for i in range(len(first_elem_node.input)):
                if first_elem_node.input[i] == T1.output[0]:
                    first_elem_node.input[i] = new_input

        # Determine the new tensor that should replace T2's output.
        # If there are elementwise nodes between, use the output of the last elementwise node.
        # Otherwise (direct T1->T2), use T1.input[0].
        if len(chain) > 2:
            new_output = chain[-2].output[0]
        else:
            new_output = T1.input[0]

        # Rewire all consumers of T2's output to use new_output.
        for n in graph.node:
            for i in range(len(n.input)):
                if n.input[i] == T2.output[0]:
                    n.input[i] = new_output
        # Also, update graph outputs if needed.
        for tensor in graph.output:
            if tensor.name == T2.output[0]:
                tensor.name = new_output

        # Mark T1 and T2 for removal.
        nodes_to_remove.extend([T1, T2])

    # Remove marked nodes.
    new_nodes = [n for n in graph.node if n not in nodes_to_remove]
    del graph.node[:]
    graph.node.extend(new_nodes)
    return onnx_model



from jax import core, numpy as jnp
from jax.extend.core import Primitive
from onnx import helper
from typing import TYPE_CHECKING, Tuple, List, Union
from jax2onnx.plugin_system import register_primitive, PrimitivePlugin

if TYPE_CHECKING:
    from jax2onnx.converter.converter import Jaxpr2OnnxConverter

# Define the MatMul primitive
jnp.matmul_p = Primitive("jnp.matmul")
jnp.matmul_p.multiple_results = False  # Correct initialization


@register_primitive(
    jaxpr_primitive=jnp.matmul_p.name,
    jax_doc="https://jax.readthedocs.io/en/latest/_autosummary/jax.numpy.matmul.html",
    onnx=[
        {
            "component": "MatMul",
            "doc": "https://onnx.ai/onnx/operators/onnx__MatMul.html",
        }
    ],
    since="v0.1.0",
    context="primitives.jnp",
    testcases=[
        {
            "testcase": "matmul_2d",
            "callable": lambda a, b: jnp.matmul(a, b),
            "input_shapes": [(3, 4), (4, 5)],
        },
        {
            "testcase": "matmul_1d_2d",
            "callable": lambda a, b: jnp.matmul(a, b),
            "input_shapes": [(4,), (4, 5)],
        },
        {
            "testcase": "matmul_2d_1d",
            "callable": lambda a, b: jnp.matmul(a, b),
            "input_shapes": [(3, 4), (4,)],
        },
        {
            "testcase": "matmul_dynamic",
            "callable": lambda a, b: jnp.matmul(a, b),
            "input_shapes": [("B", 3, 4), ("B", 4, 5)],
        },
        {
            "testcase": "matmul_dynamic_a",
            "callable": lambda a, b: jnp.matmul(a, b),
            "input_shapes": [("B", 3), (3, 4)],
        },
        {
            "testcase": "matmul_dynamic_b",
            "callable": lambda a, b: jnp.matmul(a, b),
            "input_shapes": [(3, "B"), ("B", 4)],
        },
        {
            "testcase": "matmul_1d",
            "callable": lambda a, b: jnp.matmul(a, b),
            "input_shapes": [(4,), (4,)],
        },
        {
            "testcase": "matmul_3d",
            "callable": lambda a, b: jnp.matmul(a, b),
            "input_shapes": [(2, 3, 4), (2, 4, 5)],
        },
    ],
)
class MatMulPlugin(PrimitivePlugin):
    """
    Plugin for converting jax.numpy.matmul to ONNX.
    """

    @staticmethod
    def _get_dynamic_output_shape(
        a_shape: Tuple[Union[int, str], ...], b_shape: Tuple[Union[int, str], ...]
    ) -> Tuple[Union[int, str], ...]:
        """Calculates the output shape of jnp.matmul while handling dynamic dimensions."""
        a_rank, b_rank = len(a_shape), len(b_shape)

        if a_rank == 1 and b_rank == 1:
            if (
                a_shape[0] == b_shape[0]
                or isinstance(a_shape[0], str)
                or isinstance(b_shape[0], str)
            ):
                return ()  # Scalar output
            raise ValueError("Incompatible shapes for matmul")

        a_shape_norm = a_shape if a_rank > 1 else (1,) + a_shape
        b_shape_norm = b_shape if b_rank > 1 else b_shape + (1,)

        a_rows, a_cols = a_shape_norm[-2], a_shape_norm[-1]
        b_rows, b_cols = b_shape_norm[-2], b_shape_norm[-1]

        if a_cols != b_rows and not (
            isinstance(a_cols, str) or isinstance(b_rows, str)
        ):
            raise ValueError(f"Incompatible shapes for matmul: {a_shape} and {b_shape}")

        batch_dims: List[Union[int, str]] = []
        max_rank = max(a_rank, b_rank)
        for i in range(max_rank - 2):
            a_dim = a_shape[i] if i < a_rank - 2 else 1
            b_dim = b_shape[i] if i < b_rank - 2 else 1
            if isinstance(a_dim, str) or isinstance(b_dim, str):
                batch_dims.append(
                    a_dim if isinstance(b_dim, int) and b_dim == 1 else b_dim
                )
            else:
                batch_dims.append(a_dim if a_dim != 1 else b_dim)

        output_shape = tuple(batch_dims) + (a_rows, b_cols)

        if a_rank == 1:
            output_shape = output_shape[1:]
        if b_rank == 1:
            output_shape = output_shape[:-1]

        return output_shape

    @staticmethod
    def abstract_eval(a, b):
        """Abstract evaluation function for MatMul."""
        output_shape = MatMulPlugin._get_dynamic_output_shape(a.shape, b.shape)
        return core.ShapedArray(output_shape, a.dtype)

    def to_onnx(self, s: "Jaxpr2OnnxConverter", node_inputs, node_outputs, params):
        """Handles conversion of MatMul to ONNX format."""
        input_names = [s.get_name(var) for var in node_inputs]
        output_name = s.get_name(node_outputs[0])

        input_shapes = [inp.aval.shape for inp in node_inputs]
        output_shape = MatMulPlugin._get_dynamic_output_shape(
            input_shapes[0], input_shapes[1]
        )

        matmul_node = helper.make_node(
            "MatMul",
            inputs=input_names,
            outputs=[output_name],
            name=s.get_unique_name("matmul"),
        )
        s.add_node(matmul_node)
        s.add_shape_info(output_name, output_shape)

    @staticmethod
    def _matmul(a, b):
        """Defines the primitive binding for MatMul."""
        return jnp.matmul_p.bind(a, b)

    @staticmethod
    def get_monkey_patch():
        """Provides patching information for MatMul."""

        def patched_matmul(a, b):
            return MatMulPlugin._matmul(a, b)

        return patched_matmul

    @staticmethod
    def patch_info():
        """Provides patching information for MatMul."""
        return {
            "patch_targets": [jnp],
            "patch_function": lambda _: MatMulPlugin.get_monkey_patch(),
            "target_attribute": "matmul",
        }


# Register abstract evaluation function
jnp.matmul_p.def_abstract_eval(MatMulPlugin.abstract_eval)



from jax import core, numpy as jnp
from jax.extend.core import Primitive
from onnx import helper
from typing import TYPE_CHECKING
from jax2onnx.plugin_system import register_primitive, PrimitivePlugin

if TYPE_CHECKING:
    from jax2onnx.converter.converter import Jaxpr2OnnxConverter

# Define the Concat primitive
jnp.concat_p = Primitive("jnp.concat")
jnp.concat_p.multiple_results = False  # Correct initialization


@register_primitive(
    jaxpr_primitive=jnp.concat_p.name,
    jax_doc="https://docs.jax.dev/en/latest/_autosummary/jax.numpy.concat.html",
    onnx=[
        {
            "component": "Concat",
            "doc": "https://onnx.ai/onnx/operators/onnx__Concat.html",
        }
    ],
    since="v0.1.0",
    context="primitives.jnp",
    testcases=[
        {
            "testcase": "concat",
            "callable": lambda a, b: jnp.concat((a, b), axis=0),
            "input_shapes": [(3,), (3,)],
        }
    ],
)
class ConcatPlugin(PrimitivePlugin):
    """
    Plugin for converting jax.numpy.concatenate to ONNX.  Note:  jax.numpy.concat
    is an alias for jax.numpy.concatenate.
    """

    @staticmethod
    def abstract_eval(*arrays, axis):
        """Abstract evaluation function for Concat."""
        base_shape = list(arrays[0].shape)
        total_dim = sum(a.shape[axis] for a in arrays)
        base_shape[axis] = total_dim
        return core.ShapedArray(tuple(base_shape), arrays[0].dtype)

    def to_onnx(self, s: "Jaxpr2OnnxConverter", node_inputs, node_outputs, params):
        """Handles conversion of Concat to ONNX format."""
        # Expect node_inputs: a list of arrays to concatenate.
        axis = params.get("axis", 0)
        input_names = [s.get_name(var) for var in node_inputs]
        output_name = s.get_name(node_outputs[0])

        concat_node = helper.make_node(
            "Concat",
            inputs=input_names,
            outputs=[output_name],
            name=s.get_unique_name("concat"),
            axis=axis,
        )
        s.add_node(concat_node)

    @staticmethod
    def _concat(arrays, axis):
        """Defines the primitive binding for Concat."""
        return jnp.concat_p.bind(*arrays, axis=axis)

    @staticmethod
    def get_monkey_patch():
        """Provides patching information for Concat."""

        def patched_concat(arrays, axis):
            return ConcatPlugin._concat(arrays, axis)

        return patched_concat

    @staticmethod
    def patch_info():
        """Provides patching information for Concat."""
        return {
            "patch_targets": [jnp],
            "patch_function": lambda _: ConcatPlugin.get_monkey_patch(),
            "target_attribute": "concatenate",  # Correct attribute name
        }


# Register abstract evaluation function
jnp.concat_p.def_abstract_eval(ConcatPlugin.abstract_eval)



# file: jax2onnx/plugins/flax/nnx/avg_pool.py
from jax import core
from jax.extend.core import Primitive
from flax import nnx
from onnx import helper
from typing import TYPE_CHECKING, Tuple, Sequence
from jax2onnx.plugin_system import register_primitive, PrimitivePlugin

if TYPE_CHECKING:
    from jax2onnx.converter.converter import Jaxpr2OnnxConverter

# Define the avg_pool primitive
nnx.avg_pool_p = Primitive("nnx.avg_pool")
nnx.avg_pool_p.multiple_results = False  # Correctly set at initialization


@register_primitive(
    jaxpr_primitive=nnx.avg_pool_p.name,
    jax_doc="https://flax-linen.readthedocs.io/en/latest/api_reference/flax.linen/layers.html#flax.linen.avg_pool",
    onnx=[
        {
            "component": "AveragePool",
            "doc": "https://onnx.ai/onnx/operators/onnx__AveragePool.html",
        },
        {
            "component": "Transpose",
            "doc": "https://onnx.ai/onnx/operators/onnx__Transpose.html",
        },
    ],
    since="v0.1.0",
    context="primitives.nnx",
    testcases=[
        {
            "testcase": "avg_pool",
            "callable": lambda x: nnx.avg_pool(
                x, window_shape=(2, 2), strides=(2, 2), padding="VALID"
            ),
            "input_shapes": [(1, 32, 32, 3)],
        },
        {
            "testcase": "avg_pool_same_padding",
            "callable": lambda x: nnx.avg_pool(
                x, window_shape=(2, 2), strides=(2, 2), padding="SAME"
            ),
            "input_shapes": [(1, 32, 32, 3)],
        },
        {
            "testcase": "avg_pool_default_padding",
            "callable": lambda x: nnx.avg_pool(x, window_shape=(2, 2), strides=(2, 2)),
            "input_shapes": [(1, 32, 32, 3)],
        },
    ],
)
class AvgPoolPlugin(PrimitivePlugin):
    """
    Plugin for converting flax.nnx.avg_pool to ONNX.
    """

    @staticmethod
    def _compute_avg_pool_output_shape(
        x_shape: Tuple[int, ...],
        window_shape: Sequence[int],
        strides: Sequence[int],
        padding: str,
        input_format: str = "NHWC",
    ) -> Tuple[int, ...]:
        """Computes the output shape of avg_pool operation."""
        if input_format == "NHWC":
            spatial_dims = x_shape[1:-1]  # Extract H, W from NHWC
            batch_dim = x_shape[0]
            channel_dim = x_shape[-1]
        elif input_format == "NCHW":
            spatial_dims = x_shape[2:]  # Extract H, W from NCHW
            batch_dim = x_shape[0]
            channel_dim = x_shape[1]
        else:
            raise ValueError("Invalid input_format. Must be 'NHWC' or 'NCHW'.")

        out_dims = []
        for dim, w, s in zip(spatial_dims, window_shape, strides):
            if padding.upper() == "VALID":
                out_dim = (dim - w) // s + 1
            elif padding.upper() == "SAME":
                out_dim = -(-dim // s)  # Equivalent to ceil(dim / s)
            else:
                raise ValueError("Unsupported padding: " + padding)
            out_dims.append(out_dim)

        return (
            (batch_dim,) + tuple(out_dims) + (channel_dim,)
            if input_format == "NHWC"
            else (batch_dim, channel_dim) + tuple(out_dims)
        )

    @staticmethod
    def abstract_eval(inputs, window_shape, strides, padding, count_include_pad):
        """Abstract evaluation function for avg_pool."""
        out_shape = AvgPoolPlugin._compute_avg_pool_output_shape(
            inputs.shape, window_shape, strides, padding, input_format="NHWC"
        )
        return core.ShapedArray(out_shape, inputs.dtype)

    def to_onnx(self, s: "Jaxpr2OnnxConverter", node_inputs, node_outputs, params):
        """Handles conversion of avg_pool to ONNX format."""
        input_var = node_inputs[0]
        input_name = s.get_name(input_var)
        final_output_name = s.get_name(node_outputs[0])

        window_shape = params.get("window_shape")
        strides = params.get("strides")
        padding = params.get("padding")
        count_include_pad = params.get("count_include_pad")

        jax_input_shape = input_var.aval.shape

        # === Pre-Transpose: NHWC -> NCHW ===
        pre_transpose_name = s.get_unique_name("pre_transpose")
        pre_transpose_node = helper.make_node(
            "Transpose",
            inputs=[input_name],
            outputs=[pre_transpose_name],
            name=s.get_unique_name("transpose_pre"),
            perm=[0, 3, 1, 2],  # NHWC -> NCHW
        )
        s.add_node(pre_transpose_node)
        pre_transposed_shape = (
            jax_input_shape[0],
            jax_input_shape[3],
            jax_input_shape[1],
            jax_input_shape[2],
        )
        s.add_shape_info(pre_transpose_name, pre_transposed_shape)

        # === AveragePool Node in ONNX (operates in NCHW) ===
        pool_out_name = s.get_unique_name("avg_pool_output")

        if padding.upper() == "SAME":
            pads = []
            for i in range(len(window_shape)):
                in_dim = pre_transposed_shape[2 + i]
                out_dim = -(-in_dim // strides[i])
                total_pad = max(
                    0, (out_dim - 1) * strides[i] + window_shape[i] - in_dim
                )
                pad_before = total_pad // 2
                pad_after = total_pad - pad_before
                pads.extend([pad_before, pad_after])
        else:
            pads = [0] * (2 * len(window_shape))

        # ONNX's AveragePool `count_include_pad` attribute:
        #   - 0: Exclude padding pixels from the average (what we want)
        #   - 1: Include padding pixels in the average
        onnx_count_include_pad = 0 if count_include_pad else 1

        avg_pool_node = helper.make_node(
            "AveragePool",
            inputs=[pre_transpose_name],
            outputs=[pool_out_name],
            name=s.get_unique_name("avg_pool"),
            kernel_shape=window_shape,
            strides=strides,
            pads=pads,
            count_include_pad=onnx_count_include_pad,
        )
        s.add_node(avg_pool_node)

        avgpool_output_shape_nchw = AvgPoolPlugin._compute_avg_pool_output_shape(
            pre_transposed_shape, window_shape, strides, padding, input_format="NCHW"
        )
        s.add_shape_info(pool_out_name, avgpool_output_shape_nchw)

        # === Post-Transpose: NCHW -> NHWC ===
        post_transpose_node = helper.make_node(
            "Transpose",
            inputs=[pool_out_name],
            outputs=[final_output_name],
            name=s.get_unique_name("transpose_post"),
            perm=[0, 2, 3, 1],  # NCHW -> NHWC
        )
        s.add_node(post_transpose_node)

    @staticmethod
    def _avg_pool(inputs, window_shape, strides, padding, count_include_pad):
        """Defines the primitive binding for avg_pool."""
        if strides is None:
            strides = (1, 1)
        return nnx.avg_pool_p.bind(
            inputs,
            window_shape=window_shape,
            strides=strides,
            padding=padding,
            count_include_pad=count_include_pad,
        )

    @staticmethod
    def get_monkey_patch():
        """Provides patching information for avg_pool."""

        def patched_avg_pool(
            inputs, window_shape, strides=None, padding="VALID", count_include_pad=True
        ):
            return AvgPoolPlugin._avg_pool(
                inputs, window_shape, strides, padding, count_include_pad
            )

        return patched_avg_pool

    @staticmethod
    def patch_info():
        """Provides patching information for avg_pool."""
        return {
            "patch_targets": [nnx],
            "patch_function": lambda _: AvgPoolPlugin.get_monkey_patch(),
            "target_attribute": "avg_pool",
        }


# Register abstract evaluation function.
nnx.avg_pool_p.def_abstract_eval(AvgPoolPlugin.abstract_eval)



# file: jax2onnx/examples/mnist_vit.py

import jax
import jax.numpy as jnp
from flax import nnx
from jax2onnx.plugin_system import register_example
from typing import List


# ---------------------------------------------------------------------------
# Helper for positional embeddings
# ---------------------------------------------------------------------------
def create_sinusoidal_embeddings(num_patches: int, num_hiddens: int) -> jnp.ndarray:
    position = jnp.arange(num_patches + 1)[:, jnp.newaxis]
    div_term = jnp.exp(
        jnp.arange(0, num_hiddens, 2) * -(jnp.log(10000.0) / num_hiddens)
    )
    pos_embedding = jnp.zeros((num_patches + 1, num_hiddens))
    pos_embedding = pos_embedding.at[:, 0::2].set(jnp.sin(position * div_term))
    pos_embedding = pos_embedding.at[:, 1::2].set(jnp.cos(position * div_term))
    return pos_embedding[jnp.newaxis, :, :]


# ---------------------------------------------------------------------------
# Model components
# ---------------------------------------------------------------------------
class PatchEmbedding(nnx.Module):
    """Patch embedding for Vision Transformers."""

    def __init__(
        self, height, width, patch_size, num_hiddens, in_features, *, rngs: nnx.Rngs
    ):
        num_patches_h, num_patches_w = height // patch_size, width // patch_size
        num_patches = num_patches_h * num_patches_w
        self.layers = [
            lambda x: x.reshape(
                x.shape[0],
                height // patch_size,
                patch_size,
                width // patch_size,
                patch_size,
                in_features,
            ),
            lambda x: x.transpose((0, 1, 3, 2, 4, 5)),
            lambda x: x.reshape(-1, num_patches, patch_size * patch_size * x.shape[-1]),
            nnx.Linear(
                in_features=patch_size * patch_size * in_features,
                out_features=num_hiddens,
                rngs=rngs,
            ),
        ]

    def __call__(self, x: jnp.ndarray) -> jnp.ndarray:
        for layer in self.layers:
            x = layer(x)
        return x


register_example(
    component="PatchEmbedding",
    description="Cutting the image into patches and linearly embedding them.",
    source="https://github.com/google/flax/blob/main/README.md",
    since="v0.1.0",
    context="examples.nnx",
    children=["flax.nnx.Linear", "jax.numpy.Transpose", "jax.numpy.Reshape"],
    testcases=[
        {
            "testcase": "patch_embedding",
            "callable": PatchEmbedding(
                height=28,
                width=28,
                patch_size=4,
                num_hiddens=256,
                in_features=1,
                rngs=nnx.Rngs(0),
            ),
            "input_shapes": [(3, 28, 28, 1)],
        }
    ],
)


class ConvEmbedding(nnx.Module):
    """Convolutional embedding for MNIST."""

    def __init__(
        self,
        W: int = 28,
        H: int = 28,
        embed_dims: List[int] = [32, 64, 128],
        kernel_size: int = 3,
        strides: List[int] = [1, 2, 2],
        dropout_rate: float = 0.5,
        *,
        rngs=nnx.Rngs(0),
    ):
        padding = "SAME"
        layernormfeatures = embed_dims[-1] * W // 4 * H // 4
        self.layers = [
            nnx.Conv(
                in_features=1,
                out_features=embed_dims[0],
                kernel_size=(kernel_size, kernel_size),
                strides=(strides[0], strides[0]),
                padding=padding,
                rngs=rngs,
            ),
            lambda x: nnx.gelu(x, approximate=False),
            nnx.Conv(
                in_features=embed_dims[0],
                out_features=embed_dims[1],
                kernel_size=(kernel_size, kernel_size),
                strides=(strides[1], strides[1]),
                padding=padding,
                rngs=rngs,
            ),
            lambda x: nnx.gelu(x, approximate=False),
            nnx.Conv(
                in_features=embed_dims[1],
                out_features=embed_dims[2],
                kernel_size=(kernel_size, kernel_size),
                strides=(strides[2], strides[2]),
                padding=padding,
                rngs=rngs,
            ),
            lambda x: nnx.gelu(x, approximate=False),
            nnx.LayerNorm(
                num_features=layernormfeatures,
                reduction_axes=(1, 2, 3),
                feature_axes=(1, 2, 3),
                rngs=rngs,
            ),
            nnx.Dropout(rate=dropout_rate, rngs=rngs),
        ]

    def __call__(self, x: jnp.ndarray) -> jnp.ndarray:
        for layer in self.layers:
            x = layer(x)
        B, H, W, C = x.shape
        x = x.reshape(B, H * W, C)
        return x


register_example(
    component="ConvEmbedding",
    description="Convolutional Token Embedding for MNIST with hierarchical downsampling.",
    source="https://github.com/google/flax/blob/main/README.md",
    since="v0.1.0",
    context="examples.nnx",
    children=[
        "flax.nnx.Conv",
        "flax.nnx.LayerNorm",
        "jax.numpy.Reshape",
        "jax.nn.relu",
    ],
    testcases=[
        {
            "testcase": "mnist_conv_embedding",
            "callable": ConvEmbedding(
                embed_dims=[32, 64, 128],
                kernel_size=3,
                strides=[1, 2, 2],
                rngs=nnx.Rngs(0),
            ),
            "input_shapes": [(3, 28, 28, 1)],
        }
    ],
)


class MLPBlock(nnx.Module):
    """MLP block for Transformer layers."""

    def __init__(self, num_hiddens, mlp_dim, dropout_rate=0.1, *, rngs: nnx.Rngs):
        self.layers = [
            nnx.Linear(num_hiddens, mlp_dim, rngs=rngs),
            lambda x: nnx.gelu(x, approximate=False),
            nnx.Dropout(rate=dropout_rate, rngs=rngs),
            nnx.Linear(mlp_dim, num_hiddens, rngs=rngs),
            nnx.Dropout(rate=dropout_rate, rngs=rngs),
        ]

    def __call__(self, x: jnp.ndarray, deterministic: bool = True) -> jnp.ndarray:
        for layer in self.layers:
            if isinstance(layer, nnx.Dropout):
                x = layer(x, deterministic=deterministic)
            else:
                x = layer(x)
        return x


register_example(
    component="MLPBlock",
    description="MLP in Transformer",
    source="https://github.com/google/flax/blob/main/README.md",
    since="v0.1.0",
    context="examples.nnx",
    children=["flax.nnx.Linear", "flax.nnx.Dropout", "flax.nnx.gelu"],
    testcases=[
        {
            "testcase": "mlp_block",
            "callable": MLPBlock(
                num_hiddens=256, mlp_dim=512, dropout_rate=0.1, rngs=nnx.Rngs(0)
            ),
            "input_shapes": [(1, 10, 256)],
        },
    ],
)


class TransformerBlock(nnx.Module):
    """Transformer block with multi-head attention and MLP."""

    def __init__(
        self,
        num_hiddens: int,
        num_heads: int,
        mlp_dim: int,
        attention_dropout_rate: float = 0.1,
        mlp_dropout_rate: float = 0.1,
        *,
        rngs: nnx.Rngs,
    ):
        self.rng_collection = rngs
        self.layer_norm1 = nnx.LayerNorm(num_hiddens, rngs=rngs)
        self.attention = nnx.MultiHeadAttention(
            num_heads=num_heads,
            qkv_features=num_hiddens,
            out_features=num_hiddens,
            in_features=num_hiddens,
            rngs=rngs,
            decode=False,
        )
        self.layer_norm2 = nnx.LayerNorm(num_hiddens, rngs=rngs)
        self.mlp_block = MLPBlock(num_hiddens, mlp_dim, mlp_dropout_rate, rngs=rngs)
        self.dropout = nnx.Dropout(rate=attention_dropout_rate, rngs=rngs)

    def __call__(self, x: jnp.ndarray, deterministic: bool = True) -> jnp.ndarray:
        y = self.attention(self.layer_norm1(x))
        y = self.dropout(y, deterministic=deterministic)
        x = x + y
        return x + self.mlp_block(self.layer_norm2(x), deterministic)


register_example(
    component="TransformerBlock",
    description="Transformer from 'Attention Is All You Need.'",
    source="https://github.com/google/flax/blob/main/README.md",
    since="v0.1.0",
    context="examples.nnx",
    children=[
        "flax.nnx.MultiHeadAttention",
        "flax.nnx.LayerNorm",
        "MLPBlock",
        "flax.nnx.Dropout",
    ],
    testcases=[
        {
            "testcase": "transformer_block",
            "callable": TransformerBlock(
                num_hiddens=256,
                num_heads=8,
                mlp_dim=512,
                attention_dropout_rate=0.1,
                mlp_dropout_rate=0.1,
                rngs=nnx.Rngs(0),
            ),
            "input_shapes": [(1, 10, 256)],
        },
    ],
)


class VisionTransformer(nnx.Module):
    """Vision Transformer model for MNIST with configurable embedding type."""

    def __init__(
        self,
        height: int,
        width: int,
        num_hiddens: int,
        num_layers: int,
        num_heads: int,
        mlp_dim: int,
        num_classes: int,
        embed_dims: List[int] = [32, 128, 256],
        kernel_size: int = 3,
        strides: List[int] = [1, 2, 2],
        patch_size: int = 4,
        embedding_type: str = "conv",  # "conv" or "patch"
        embedding_dropout_rate: float = 0.1,
        attention_dropout_rate: float = 0.1,
        mlp_dropout_rate: float = 0.1,
        *,
        rngs: nnx.Rngs,
    ):
        if embedding_type not in ["conv", "patch"]:
            raise ValueError("embedding_type must be either 'conv' or 'patch'")
        if embedding_type == "conv":
            if len(embed_dims) != 3 or embed_dims[2] != num_hiddens:
                raise ValueError(
                    "embed_dims should be a list of size 3 with embed_dims[2] == num_hiddens"
                )
            self.embedding = ConvEmbedding(
                embed_dims=embed_dims,
                kernel_size=kernel_size,
                strides=strides,
                dropout_rate=embedding_dropout_rate,
                rngs=rngs,
            )
            num_patches = (height // 4) * (width // 4)
        else:
            self.embedding = PatchEmbedding(
                height=height,
                width=width,
                patch_size=patch_size,
                num_hiddens=num_hiddens,
                in_features=1,
                rngs=rngs,
            )
            num_patches = (height // patch_size) * (width // patch_size)
        self.cls_token = nnx.Param(
            jax.random.normal(rngs.params(), (1, 1, num_hiddens))
        )
        self.positional_embedding = nnx.Param(
            create_sinusoidal_embeddings(num_patches, num_hiddens)
        )
        self.transformer_blocks = [
            TransformerBlock(
                num_hiddens,
                num_heads,
                mlp_dim,
                attention_dropout_rate,
                mlp_dropout_rate,
                rngs=rngs,
            )
            for _ in range(num_layers)
        ]
        self.layer_norm = nnx.LayerNorm(num_features=num_hiddens, rngs=rngs)
        self.dense = nnx.Linear(num_hiddens, num_classes, rngs=rngs)

    def __call__(self, x: jnp.ndarray, deterministic: bool = True) -> jnp.ndarray:
        x = self.embedding(x)
        batch_size = x.shape[0]
        cls_tokens = jnp.tile(self.cls_token.value, (batch_size, 1, 1))
        x = jnp.concatenate([cls_tokens, x], axis=1)
        pos_emb_expanded = jax.lax.dynamic_slice(
            self.positional_embedding.value, (0, 0, 0), (1, x.shape[1], x.shape[2])
        )
        pos_emb_expanded = jnp.asarray(pos_emb_expanded)
        x = x + pos_emb_expanded
        for block in self.transformer_blocks:
            x = block(x, deterministic)
        x = self.layer_norm(x)
        x = x[:, 0, :]
        return nnx.log_softmax(self.dense(x))


register_example(
    component="VisionTransformer",
    description="A Vision Transformer (ViT) model for MNIST with configurable embedding type.",
    source="https://github.com/google/flax/blob/main/README.md",
    since="v0.2.0",
    context="examples.nnx",
    children=[
        "PatchEmbedding",
        "ConvEmbedding",
        "MLPBlock",
        "TransformerBlock",
        "nnx.MultiHeadAttention",
        "nnx.LayerNorm",
        "nnx.Linear",
        "nnx.gelu",
        "nnx.Dropout",
        "nnx.Param",
    ],
    testcases=[
        {
            "testcase": "vit_conv_embedding",
            "callable": VisionTransformer(
                height=28,
                width=28,
                num_hiddens=256,
                num_layers=6,
                num_heads=8,
                mlp_dim=512,
                num_classes=10,
                embedding_type="conv",
                rngs=nnx.Rngs(0),
            ),
            "input_shapes": [(2, 28, 28, 1)],
        },
        {
            "testcase": "vit_patch_embedding",
            "callable": VisionTransformer(
                height=28,
                width=28,
                num_hiddens=256,
                num_layers=6,
                num_heads=8,
                mlp_dim=512,
                num_classes=10,
                embedding_type="patch",
                patch_size=4,
                rngs=nnx.Rngs(0),
            ),
            "input_shapes": [(2, 28, 28, 1)],
        },
    ],
)



# file: jax2onnx/examples/multi_head_attention.py

from flax import nnx
from jax2onnx.plugin_system import register_example


register_example(
    component="MultiHeadAttention",
    description="This is a multi-head attention module implemented by Flax/nnx that has no ONNX correspondent on the same granularity.",
    source="https://github.com/google/flax/blob/main/README.md",
    since="v0.2.0",
    context="examples.nnx",
    children=["nnx.GeneralLinear", "nnx.dot_product_attention"],
    testcases=[
        {
            "testcase": "multihead_attention",
            "callable": nnx.MultiHeadAttention(
                num_heads=8,
                in_features=256,
                qkv_features=256,
                out_features=256,
                rngs=nnx.Rngs(0),
                decode=False,
            ),
            "input_shapes": [(2, 4, 256)],
        }
    ],
)



from jax import core
from jax.extend.core import Primitive
from flax import nnx
from onnx import helper
from typing import TYPE_CHECKING
from jax2onnx.plugin_system import register_primitive, PrimitivePlugin

if TYPE_CHECKING:
    from jax2onnx.converter.converter import Jaxpr2OnnxConverter

# Define the ELU primitive
nnx.elu_p = Primitive("nnx.elu")
nnx.elu_p.multiple_results = False  # Correctly set at initialization


@register_primitive(
    jaxpr_primitive=nnx.elu_p.name,
    jax_doc="https://jax.readthedocs.io/en/latest/_autosummary/jax.nn.elu.html",
    onnx=[
        {
            "component": "Elu",
            "doc": "https://onnx.ai/onnx/operators/onnx__Elu.html",
        }
    ],
    since="v0.1.0",
    context="primitives.nnx",
    testcases=[
        {
            "testcase": "elu",
            "callable": lambda x: nnx.elu(x),
            "input_shapes": [(3,)],
        }
    ],
)
class EluPlugin(PrimitivePlugin):
    """
    Plugin for converting flax.nnx.elu to ONNX.
    """

    @staticmethod
    def abstract_eval(x, alpha=1.0):
        """Abstract evaluation function for ELU."""
        return core.ShapedArray(x.shape, x.dtype)

    def to_onnx(self, s: "Jaxpr2OnnxConverter", node_inputs, node_outputs, params):
        """Handles conversion of ELU to ONNX format."""
        input_var = node_inputs[0]
        output_var = node_outputs[0]

        input_name = s.get_name(input_var)
        output_name = s.get_name(output_var)

        # Retrieve the alpha parameter (defaulting to 1.0 if not provided)
        alpha = params.get("alpha", 1.0)

        elu_node = helper.make_node(
            "Elu",
            inputs=[input_name],
            outputs=[output_name],
            name=s.get_unique_name("elu"),
            alpha=alpha,
        )
        s.add_node(elu_node)

    @staticmethod
    def _elu(x, alpha=1.0):
        """Defines the primitive binding for ELU."""
        return nnx.elu_p.bind(x, alpha=alpha)

    @staticmethod
    def get_monkey_patch():
        """Provides patching information for ELU."""

        def patched_elu(x, alpha=1.0):
            return EluPlugin._elu(x, alpha)

        return patched_elu

    @staticmethod
    def patch_info():
        """Provides patching information for ELU."""
        return {
            "patch_targets": [nnx],
            "patch_function": lambda _: EluPlugin.get_monkey_patch(),
            "target_attribute": "elu",
        }


# Register abstract evaluation function
nnx.elu_p.def_abstract_eval(EluPlugin.abstract_eval)



from jax import core
from jax.extend.core import Primitive
from flax import nnx
from onnx import helper
from typing import TYPE_CHECKING
from jax2onnx.plugin_system import register_primitive, PrimitivePlugin

if TYPE_CHECKING:
    from jax2onnx.converter.converter import Jaxpr2OnnxConverter

# Define the Tanh primitive
nnx.tanh_p = Primitive("nnx.tanh")
nnx.tanh_p.multiple_results = False  # Correct initialization


@register_primitive(
    jaxpr_primitive=nnx.tanh_p.name,
    jax_doc="https://flax.readthedocs.io/en/latest/api_reference/flax.nnx/nn/activations.html#flax.nnx.tanh",
    onnx=[
        {
            "component": "Tanh",
            "doc": "https://onnx.ai/onnx/operators/onnx__Tanh.html",
        }
    ],
    since="v0.1.0",
    context="primitives.nnx",
    testcases=[
        {
            "testcase": "tanh",
            "callable": lambda x: nnx.tanh(x),
            "input_shapes": [(3,)],
        }
    ],
)
class TanhPlugin(PrimitivePlugin):
    """
    Plugin for converting flax.nnx.tanh to ONNX.
    """

    @staticmethod
    def abstract_eval(x):
        """Abstract evaluation function for Tanh."""
        return core.ShapedArray(x.shape, x.dtype)

    def to_onnx(self, s: "Jaxpr2OnnxConverter", node_inputs, node_outputs, params):
        """Handles conversion of Tanh to ONNX format."""
        input_var = node_inputs[0]
        output_var = node_outputs[0]

        input_name = s.get_name(input_var)
        output_name = s.get_name(output_var)

        tanh_node = helper.make_node(
            "Tanh",
            inputs=[input_name],
            outputs=[output_name],
            name=s.get_unique_name("tanh"),
        )
        s.add_node(tanh_node)

    @staticmethod
    def _tanh(x):
        """Defines the primitive binding for Tanh."""
        return nnx.tanh_p.bind(x)

    @staticmethod
    def get_monkey_patch():
        """Provides patching information for Tanh."""

        def patched_tanh(x):
            return TanhPlugin._tanh(x)

        return patched_tanh

    @staticmethod
    def patch_info():
        """Provides patching information for Tanh."""
        return {
            "patch_targets": [nnx],
            "patch_function": lambda _: TanhPlugin.get_monkey_patch(),
            "target_attribute": "tanh",
        }


# Register abstract evaluation function
nnx.tanh_p.def_abstract_eval(TanhPlugin.abstract_eval)



from jax import core
from jax.extend.core import Primitive
from flax import nnx
from onnx import helper
from typing import TYPE_CHECKING
from jax2onnx.plugin_system import register_primitive, PrimitivePlugin

if TYPE_CHECKING:
    from jax2onnx.converter.converter import Jaxpr2OnnxConverter

# Define the Softplus primitive
nnx.softplus_p = Primitive("nnx.softplus")
nnx.softplus_p.multiple_results = False  # Correct initialization


@register_primitive(
    jaxpr_primitive=nnx.softplus_p.name,
    jax_doc="https://jax.readthedocs.io/en/latest/_autosummary/jax.nn.softplus.html",
    onnx=[
        {
            "component": "Softplus",
            "doc": "https://onnx.ai/onnx/operators/onnx__Softplus.html",
        }
    ],
    since="v0.1.0",
    context="primitives.nnx",
    testcases=[
        {
            "testcase": "softplus",
            "callable": lambda x: nnx.softplus(x),
            "input_shapes": [(3,)],
        }
    ],
)
class SoftplusPlugin(PrimitivePlugin):
    """
    Plugin for converting flax.nnx.softplus to ONNX.
    """

    @staticmethod
    def abstract_eval(x):
        """Abstract evaluation function for Softplus."""
        return core.ShapedArray(x.shape, x.dtype)

    def to_onnx(self, s: "Jaxpr2OnnxConverter", node_inputs, node_outputs, params):
        """Handles conversion of Softplus to ONNX format."""
        input_var = node_inputs[0]
        output_var = node_outputs[0]

        input_name = s.get_name(input_var)
        output_name = s.get_name(output_var)

        softplus_node = helper.make_node(
            "Softplus",
            inputs=[input_name],
            outputs=[output_name],
            name=s.get_unique_name("softplus"),
        )
        s.add_node(softplus_node)

    @staticmethod
    def _softplus(x):
        """Defines the primitive binding for Softplus."""
        return nnx.softplus_p.bind(x)

    @staticmethod
    def get_monkey_patch():
        """Provides patching information for Softplus."""

        def patched_softplus(x):
            return SoftplusPlugin._softplus(x)

        return patched_softplus

    @staticmethod
    def patch_info():
        """Provides patching information for Softplus."""
        return {
            "patch_targets": [nnx],
            "patch_function": lambda _: SoftplusPlugin.get_monkey_patch(),
            "target_attribute": "softplus",
        }


# Register abstract evaluation function
nnx.softplus_p.def_abstract_eval(SoftplusPlugin.abstract_eval)



# file: jax2onnx/examples/mlp.py

from flax import nnx
import jax
from flax.nnx import Linear, Dropout, BatchNorm
from jax2onnx.plugin_system import register_example


class MLP(nnx.Module):
    def __init__(self, din: int, dmid: int, dout: int, *, rngs: nnx.Rngs):
        self.linear1 = Linear(din, dmid, rngs=rngs)
        self.dropout = Dropout(rate=0.1, rngs=rngs)
        self.bn = BatchNorm(dmid, rngs=rngs)
        self.linear2 = Linear(dmid, dout, rngs=rngs)

    def __call__(self, x: jax.Array):
        x = nnx.gelu(self.dropout(self.bn(self.linear1(x))))
        return self.linear2(x)


register_example(
    component="MLP",
    description="A simple Multi-Layer Perceptron (MLP) with BatchNorm, Dropout, and GELU activation.",
    source="https://github.com/google/flax/blob/main/README.md",
    since="v0.1.0",
    context="examples.nnx",
    children=["nnx.Linear", "nnx.Dropout", "nnx.BatchNorm", "nnx.gelu"],
    testcases=[
        {
            "testcase": "mlp",
            "callable": MLP(din=30, dmid=20, dout=10, rngs=nnx.Rngs(17)),
            "input_shapes": [("B", 30)],
        }
    ],
)
